import configparser
import logging
from pathlib import Path

from logging_spinner import SpinnerHandler


def load_config():
    """Creates config-parser, reading the config and returning the read config

    Returns:
        Parsed config
    """
    config = configparser.ConfigParser()
    config.read(PACKAGE_ROOT_PATH / 'config.ini')
    return config


def setup_spin_logging():
    logger = logging.getLogger('archiTop_spin')
    logger.setLevel(load_config()['APP']['LOG_LEVEL'])
    logger.addHandler(SpinnerHandler())


def get_spin_logger(name: str):
    return logging.getLogger('archiTop_spin.' + name)


PACKAGE_ROOT_PATH = Path(__file__).parent
setup_spin_logging()



from dataclasses import dataclass
from functools import reduce
from typing import List

from archiTop.scryfall.scryfall_loader import load_scryfall_id_index


class ScryfallCard:
    """Class containing information for scryfall card object."""
    related_cards = set()

    def __init__(self,
                 scryfall_card_data: dict,
                 resolve_tokens=True,
                 quantity=1,
                 commander=False,
                 card_side=0):
        self.id_index = load_scryfall_id_index()

        self.commander = commander
        self.quantity = quantity
        self.name = scryfall_card_data['name']
        self.type_line = scryfall_card_data['type_line']
        self.cmc = scryfall_card_data['cmc']

        cmc_string = f'CMC{int(self.cmc)}'
        self.tabletop_name = f'{self.name} - {cmc_string}\n' \
                             f'{self.type_line}'

        self.id = scryfall_card_data['id']

        if 'image_uris' in scryfall_card_data:
            # extract image_url, choosing high rez if available
            image_uris = scryfall_card_data['image_uris']

        elif 'card_faces' in scryfall_card_data:  # card is a double sized card
            image_uris = scryfall_card_data['card_faces'][card_side]['image_uris']

            if card_side == 0:
                self.related_cards = {ScryfallCard(self.id_index[scryfall_card_data['id']],
                                                   resolve_tokens=False,
                                                   card_side=1)}

        else:
            raise Exception('Unknown card-type encountered')

        if 'large' in image_uris:
            self.image_url = image_uris['large']
        else:
            self.image_url = image_uris['normal']

        if resolve_tokens:
            related_objects = scryfall_card_data.get('all_parts', ())

            related_tokens = list(
                    filter(lambda related_object: related_object['component'] == 'token',
                           related_objects))
            related_meld_cards = list(
                    filter(lambda related_object: related_object['component'] == 'meld_result',
                           related_objects))

            related_ids = set([related_object['id'] for related_object in
                               related_tokens + related_meld_cards])

            self.related_cards |= {ScryfallCard(self.id_index[related_id],
                                                resolve_tokens=False)
                                   for related_id in related_ids}

    def __repr__(self):
        commander_identifier = '[Commander]' if self.commander else ''
        return f'Card{commander_identifier}({self.quantity: <2} x {self.name})'

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, ScryfallCard):
            return self.id == other.id

        return False

    def __hash__(self):
        return hash(self.id)


@dataclass
class ScryfallDeck:
    """Class containing information for mtg deck of cards.
        This consists of the mainboard, sideboard and token / emblems."""
    mainboard: List[ScryfallCard]  # list containing cards in mainboard
    related_cards: List[ScryfallCard]  # list containing related cards (tokens, ...)
    name: str  # name of deck
    thumbnail: bytes  # image bytes for deck thumbnail

    def __repr__(self):
        total_count = reduce(lambda a, b: a + b.quantity, self.mainboard, 0)
        return (f'Deck[{self.name}](\n'
                f'- {total_count: <3} total cards\n'
                f'- {len(self.mainboard): <3} unique cards\n'
                f'- {len(self.related_cards): <3} related cards)')



"""Sourcefile containing class to interact with archidekt services"""
from typing import List, Set

import requests

from archiTop.base_classes import DeckFetcher, DeckFetcherError
from archiTop.data_types import RawCard


class ArchidektFetcher(DeckFetcher):
    """ArchidektFetcher class, implementing abstract baseclass DeckFetcher"""
    base_url = 'https://archidekt.com/api/decks/%s/small/'

    def _parse_single_card(self, card: dict) -> RawCard:
        """Parses single card information from deck service into Card object.

        Args:
            card:   Card json object to parse information from

        Returns:
            Card class containing parsed information from card json object
        """
        card_data = card['card']

        name = card_data['oracleCard']['name']
        uid = card_data['uid']
        quantity = card['quantity']

        edition_code = card_data['edition']['editioncode']
        commander = card['category'] == 'Commander' or 'Commander' in card.get('categories', ())

        return RawCard(name, quantity, uid, edition_code, commander)

    @staticmethod
    def _handle_raw_deck_request(response: requests.Response):
        """Handles response from archidekt api, validating request was successful.

        Raises:
            DeckFetcherError: When invalid status code was encountered in response from api

        Args:
            response:   Response object from archidekt api call
        """
        try:
            response.raise_for_status()

        except requests.HTTPError as e:
            try:
                error_message = response.json()['error']

            except Exception:
                error_message = e

            raise DeckFetcherError(f'Failed to fetch archidekt deck with error:\n{error_message}')

    @staticmethod
    def _parse_card_data(raw_deck_data: dict) -> List[dict]:
        """Parses card information from deck data fetched by `_get_raw_deck_data()`.

        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            List of card json objects contained in deck
        """
        return raw_deck_data['cards']

    @staticmethod
    def _parse_deck_name(raw_deck_data: dict) -> str:
        """Parses deck name from deck data fetched by `_get_raw_deck_data()`.

        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            Name of deck
        """
        return raw_deck_data['name']

    @staticmethod
    def _parse_deck_thumbnail_url(raw_deck_data: dict) -> str:
        """Parses thumbnail url from deck data fetched by `_get_raw_deck_data()`.
        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            Thumbnail url for fetched deck information
        """
        return raw_deck_data['featured']

    @staticmethod
    def _parse_mainboard_identifier(raw_deck_data: dict) -> Set[str]:
        """Extracts valid categories for cards belonging to mainboard.
        Archidekt has functionality of marking categories as not included in the deck,
        these are being filtered out.

        Args:
            raw_deck_data:      Raw json response data from archidekt api

        Returns:
            Set containing valid categories
        """
        valid_categories = {category['name'] for category in raw_deck_data['categories']
                            if category['includedInDeck']}
        return valid_categories - {'Sideboard'}

    @staticmethod
    def _validate_single_card_mainboard(card: dict, mainboard_identifier: Set[str]) -> bool:
        """Validates whether a single card belongs to mainboard.

        Args:
            card:                   Card json object contained in fetched deck information
            mainboard_identifier:   List holding valid mainboard categories

        Returns:
            True when card is contained in mainboard, False otherwise
        """
        # return category_check and categories_checks
        category = card['categories'][0] if len(card['categories']) > 0 else None

        return category in mainboard_identifier



"""Templates for TableTop deck creation"""

# template for the final json to be imported into TableTop
final_deck_template = {
    "ObjectStates": []  # <-- card_deck_templates go here
}

# template for each deck of cards in the TableTop asset (I.e: MainBoard, SideBoard, Tokens)
card_deck_template = {
    "Name": "DeckCustom",
    "ContainedObjects": [],  # <-- card_templates go here
    "DeckIDs": [],  # <-- all cardID's from card_templates go here in order
    "CustomDeck": {},  # <-- card_asset_templates go here.
    # Notation is {"1": <card_asset_template>, "2": <>...}, same ordering as card_templates
    "Transform": {
        "posX": 0,
        "posY": 1,
        "posZ": 0,
        "rotX": 0,
        "rotY": 0,
        "rotZ": 180,
        "scaleX": 1,
        "scaleY": 1,
        "scaleZ": 1
    }
}

# template for each card in the deck
# cardID should be ascending
# cards with quantity > 1 require multiple entries with the same card id
card_template = {
    "Name": "Card",
    "CardID": None,
    "Nickname": None,
    "Transform": {
        "posX": 0,
        "posY": 0,
        "posZ": 0,
        "rotX": 0,
        "rotY": 0,
        "rotZ": 180,
        "scaleX": 1,
        "scaleY": 1,
        "scaleZ": 1
    }
}

# template for card assets (back and front image)
# one entry for each unique card
card_asset_template = {
    "FaceURL": None,
    "BackURL": None,
    "NumHeight": 1,
    "NumWidth": 1,
    "BackIsHidden": True
}



"""Sourcefile containing base_class for TableTop deck construction"""
from abc import ABC, abstractmethod
from typing import List

from archiTop.config import load_config
from archiTop.scryfall.data_types import ScryfallCard

config = load_config()


class DeckBuilder(ABC):
    """Abstract baseclass to construct TableTop deck assets"""

    current_card_id = 100
    current_card_asset_id = 1

    def __init__(self,
                 cards: List[ScryfallCard],
                 hidden=True,
                 custom_back_url: str = None):
        """Initializes deck builder with list of cards.

        Args:
            cards:      Cards to include in deck of cards
            hidden:     Determines if deck of cards is facing back up
                        Defaults to true
        """
        self.cards = list(cards)
        self.hidden = hidden

        self.card_back_url = custom_back_url or config['DECK']['DEFAULT_CARDBACK_URL']

    @abstractmethod
    def create_deck(self) -> dict:
        """Abstractmethod to be implemented by child class.
        Create the json structure for the deck of cards.

        Returns:
            TableTop card deck json structure
        """
        raise NotImplemented



"""Sourcefile containing functionality to load fetched card information from scryfall"""
import pickle
from functools import lru_cache
from pathlib import Path
from typing import Dict

from archiTop.scryfall import conf, spin_logger, resources_path


def _get_data_path() -> Path:
    if scryfall_data := list(resources_path.glob(conf['BULK_DATA_FILE_PATTERN'])):
        return scryfall_data[0]


@lru_cache
def _load_scryfall_data(file_path: Path):
    spin_logger.debug('Loading scryfall data', extra={'user_waiting': True})
    data = pickle.load(file_path.open('rb'))
    spin_logger.debug('Loaded scryfall data', extra={'user_waiting': False})
    return data


@lru_cache
def _load_scryfall_id_index(file_path: Path) -> Dict[str, Dict]:
    data = _load_scryfall_data(file_path)
    return {entry['id']: entry for entry in data}


def load_scryfall_id_index() -> Dict[str, Dict]:
    file_path = _get_data_path()
    return _load_scryfall_id_index(file_path)



"""Sourcefile containing deck builder class for decks with one card"""
from copy import deepcopy

from archiTop.base_classes import DeckBuilder
from archiTop.resources import card_asset_template, card_template


class SingleCardDeckBuilder(DeckBuilder):
    """SingleCardDeckBuilder class implementing abstract DeckBuilder class.
    Used for card decks with single card."""

    def __init__(self, *args):
        self.card_json = deepcopy(card_template)
        super().__init__(*args)

    def __repr__(self):
        return f'CardDeck({self.cards[0]})'

    def create_deck(self) -> dict:
        """Create the json structure for the card deck containing a single card.

        Returns:
            TableTop card deck json containing only a single card
        """
        card = self.cards[0]

        self.card_json['CardID'] = self.current_card_id
        self.card_json['Nickname'] = card.tabletop_name

        card_asset_json = deepcopy(card_asset_template)
        card_asset_json['FaceURL'] = card.image_url
        card_asset_json['BackURL'] = self.card_back_url
        card_asset_json['BackIsHidden'] = False

        self.card_json['CustomDeck'] = {"1": card_asset_json}
        self.card_json['Transform']['rotZ'] = 180 if self.hidden else 0

        return self.card_json



"""Sourcefile containing functionality to construct scryfall card deck"""
from functools import reduce
from typing import List, Tuple

from archiTop.data_types import RawCard, RawDeck
from archiTop.scryfall.data_types import ScryfallCard, ScryfallDeck
from archiTop.scryfall.scryfall_fetcher import syncronize_scryfall_data
from archiTop.scryfall.scryfall_loader import load_scryfall_id_index


class ScryfallDeckBuilder:

    def __init__(self, raw_deck: RawDeck):
        syncronize_scryfall_data()

        # load index to search by name and scryfall id
        self.scryfall_id_index = load_scryfall_id_index()
        self.raw_deck = raw_deck

    def construct_deck(self) -> ScryfallDeck:
        cards, tokens = self._get_scryfall_cards_for_deck()

        return ScryfallDeck(cards, tokens, self.raw_deck.name, self.raw_deck.thumbnail)

    def _get_scryfall_cards_for_deck(self) -> Tuple[List[ScryfallCard], List[ScryfallCard]]:
        mainboard = self.raw_deck.mainboard

        # convert raw cards to scryfall enriched cards
        scryfall_cards = [self._create_scryfall_card(card) for card in mainboard]

        # extract related cards for scryfall cards (token, etc.)
        related_cards = reduce(lambda set_1, set_2: set_1 | set_2,
                               [card.related_cards for card in scryfall_cards])
        return scryfall_cards, related_cards

    def _create_scryfall_card(self, card: RawCard) -> ScryfallCard:
        scryfall_data = self.scryfall_id_index[card.uid]
        return ScryfallCard(scryfall_data,
                            quantity=card.quantity,
                            commander=card.commander)



"""Sourcefile containing class fetching deck information"""
from abc import ABC, abstractmethod
from functools import reduce
from typing import Any, List

import requests

from archiTop.config import get_spin_logger
from archiTop.data_types import RawCard, RawDeck

spin_logger = get_spin_logger(__name__)


class DeckFetcherError(Exception):
    """Exception to raise when an error was encountered during deck fetching"""
    pass


class DeckFetcher(ABC):
    """Abstract baseclass to for deck fetcher"""
    base_url = None
    mainboard_cards = []

    def __init__(self, deck_id: int):
        """Initializes deck fetcher with id of deck.

        Args:
            deck_id:    DeckID to fetch deck information for
        """
        self.deck_id = deck_id

    def __repr__(self) -> str:
        total_count = reduce(lambda a, b: a + b.quantity, self.mainboard_cards, 0)
        return f'DeckFetcher({total_count} total cards, {len(self.mainboard_cards)} unique cards)'

    def _request_raw_deck(self) -> requests.Response:
        """Fetch the deck information by querying base_url combined with the deckID.

        Returns:
            Deck information in json format
        """
        spin_logger.debug('Downloading deck <%s>', self.deck_id, extra={'user_waiting': True})
        response = requests.get(self.base_url % self.deck_id)
        spin_logger.debug('Downloaded deck <%s>', self.deck_id, extra={'user_waiting': False})
        return response

    @staticmethod
    def _parse_raw_deck_data(request: requests.Response) -> dict:
        return request.json()

    def get_deck(self) -> RawDeck:
        """Fetches and parses deck information.

        Returns:
            Deck of cards, containing deck information fetched
        """
        raw_deck_response = self._request_raw_deck()

        self._handle_raw_deck_request(raw_deck_response)

        raw_deck_data = self._parse_raw_deck_data(raw_deck_response)

        deck_name = self._parse_deck_name(raw_deck_data)
        thumbnail_url = self._parse_deck_thumbnail_url(raw_deck_data)

        thumbnail = requests.get(thumbnail_url).content

        mainboard_identifier = self._parse_mainboard_identifier(raw_deck_data)
        filtered_mainboard_card_data = [
                card for card in self._parse_card_data(raw_deck_data)
                if self._validate_single_card_mainboard(card, mainboard_identifier)]

        self.mainboard_cards = [self._parse_single_card(card) for card in
                                filtered_mainboard_card_data]

        return RawDeck(self.mainboard_cards, deck_name, thumbnail)

    @abstractmethod
    def _parse_single_card(self, card: dict) -> RawCard:
        """Abstractmethod to be implemented by child class.
        Parses single card information from deck service into Card object.

        Args:
            card:   Card json object to parse information from

        Returns:
            Card class containing parsed information from card json object
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _handle_raw_deck_request(response: requests.Response):
        """Abstractmethod to be implemented by child class.
        Validates whether request to server was successful.

        Args:
            response:   Response from server request
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _parse_card_data(raw_deck_data: dict) -> List[dict]:
        """Abstractmethod to be implemented by child class.
        Parses card information from deck data fetched by `_get_raw_deck_data()`.

        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            List of card json objects contained in deck
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _parse_deck_name(raw_deck_data: dict) -> str:
        """Abstractmethod to be implemented by child class.
        Parses deck name from deck data fetched by `_get_raw_deck_data()`.

        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            Name of deck
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _parse_deck_thumbnail_url(raw_deck_data: dict) -> str:
        """Abstractmethod to be implemented by child class.
        Parses thumbnail url from deck data fetched by `_get_raw_deck_data()`.

        Args:
            raw_deck_data:  Raw server data fetched by deck data request

        Returns:
            Thumbnail url for fetched deck information
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _validate_single_card_mainboard(card: dict, mainboard_identifier: Any) -> bool:
        """Abstractmethod to be implemented by child class.
        Validates whether a single card belongs to mainboard using the passed mainboard_identifier.

        Args:
            card:                   Card json object contained in fetched deck information
            mainboard_identifier:   Identifier to validate card belongs to mainboard

        Returns:
            True when card is contained in mainboard, False otherwise
        """
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def _parse_mainboard_identifier(raw_deck_data: dict) -> Any:
        """Abstractmethod to be implemented by child class.
        Parses the identifier for mainboard cards from raw data fetched.

        Args:
            raw_deck_data:      Raw data fetched from server

        Returns:
            Identifier object
        """
        raise NotImplemented



from dataclasses import dataclass
from typing import List


@dataclass
class RawCard:
    """Class containing information for mtg card.
    One class instance can hold multiple quantities of a single card."""
    name: str                   # unique card name
    quantity: int               # quantity for card
    uid: str                    # unique card identifier
    editioncode: str = None     # edition code for card
    commander: bool = False     # flag whether card is commander

    def __repr__(self):
        return f'RawCard({self.quantity: <3}x {self.name} - {self.editioncode} - {self.uid})'


@dataclass
class RawDeck:
    """Class containing information for mtg deck of cards.
        This consists of the mainboard, sideboard and token / emblems."""
    mainboard: List[RawCard]  # list containing cards in mainboard
    name: str  # name of deck
    thumbnail: bytes  # image bytes for deck thumbnail

    def __repr__(self):
        total_cards = sum([card.quantity for card in self.mainboard])
        return f'RawDeck({total_cards} total cards, {len(self.mainboard)} unique cards)'



from setuptools import setup, find_packages

setup(
        name='ArchiTop',
        version='0.2.4',
        author='Julian Brendel',
        author_email='julian.brendel@t-online.de',
        packages=find_packages(),
        url='https://github.com/Julian-Brendel/archiTop',
        license='LICENSE',
        description='Archidekt to TableTop export cli',
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        python_requires='>=3.8',
        install_requires=[
                "requests >= 2.24",
                "logging-spinner >= 0.2.2"
        ],
        entry_points={
                'console_scripts': ['archiTop=archiTop.__main__:main',
                                    'architop=archiTop.__main__:main']
        },
        include_package_data=True
)



from pathlib import Path

from archiTop.config import PACKAGE_ROOT_PATH, get_spin_logger, load_config

spin_logger = get_spin_logger(__name__)
conf = load_config()['SCRYFALL']
resources_path = Path(PACKAGE_ROOT_PATH, 'resources')

from .scryfall_builder import ScryfallDeckBuilder
from .scryfall_fetcher import syncronize_scryfall_data, syncronize_scryfall_data
from .scryfall_loader import load_scryfall_id_index



"""Entry-point for application"""
import argparse

from archiTop.scryfall import ScryfallDeckBuilder
from archiTop.deck_builder import DeckBuilderWrapper
from archiTop.deck_fetcher import ArchidektFetcher
from archiTop.config import get_spin_logger

spin_logger = get_spin_logger(__name__)


def setup_argparse():
    parser = argparse.ArgumentParser(description='Convert archidekt deck to TableTop')
    parser.add_argument('deckID',
                        help='Archidekt deck-ID to convert')
    parser.add_argument('-n', '--name', type=str,
                        help='Optional deckname to overwrite the archidekt deckname')
    parser.add_argument('-c', '--custom_back_url', type=str,
                        help='Use custom card-back image url')
    parser.add_argument('-p', '--path', type=str,
                        help='Path to write output to')
    return parser.parse_args()


def main():
    args = setup_argparse()

    # fetch raw deck information (card names and count)
    deck = ArchidektFetcher(args.deckID).get_deck()

    # overwrite deckname if optional argument is specified
    deck.name = args.name if args.name else deck.name

    # enrich deck information with scryfall data (cmc, type_lines etc.)
    scryfall_deck = ScryfallDeckBuilder(deck).construct_deck()

    builder = DeckBuilderWrapper(scryfall_deck, custom_back_url=args.custom_back_url)
    builder.construct_final_deck()
    builder.save_deck(args.path)


if __name__ == '__main__':
    main()
